##
#
# egghttp2.tcl v1.1.0 - by strikelight ([sL] @ EFNet) (05/14/04)
#              v2.0.1 - modified by De Kus <DeKus|gmx.net> (06/09/09)
#
# Contact:
# - E-Mail: strikelight@tclscript.com
# - WWW   : http://www.TCLScript.com
# - IRC   : #Scripting @ EFNet
#
# - E-Mail: DeKus@gmx.net
# - WWW   : http://forum.egghelp.org/
# - IRC   : De_Kus, StarZ|De_Kus or DeKus on various Networks
#
##
#
# Note about version 2.x:
#
# It intends to fix problem occoured with the older version and try to parse
# the webservers return more HTTP strict. It is supposed to be backwards
# compatible with scripts written for 1.1.0, but it might also break. Feedback
# is welcome on that one.
#
# Description:
#
# This is a TCL for other scripters to use for true asynchronous
# webpage connections.
#
# I noticed the need when using the http package for tcl,
# and it would not, for some reason or other, properly
# use asynchronous connections or not do anything at all when
# trying to use async connections.
# ^- As it turns out, eggdrop1.1.5 (and I believe 1.3.x) does
#    not have Tcl_DoOneEvent in the source, so the http package fails
#    for async connections, thus the need for this script.
#
# Realizing eggdrop already had the ability to make async connections,
# I created this considerably smaller tcl (in comparison to the http
# package).
#
# So, no more fighting with the http package for async connections,
# and no more freezes when trying to connect to a page. Enjoy!
#
##
#
# History:
#
# (09/27/06) - v2.1.0 - Rewrote egghttp:control to read header diffrent than body
#                     - compressed data is read as binary
# (09/26/06) - v2.0.4 - Unsuccessfull try to get binary data parsing to work
# (09/14/06) - v2.0.3 - Added -encoding and draft of gzip and deflate handling
# (09/10/06) - v2.0.2 - Added -myaddr
#                       Removed fileevent writable
#                       Optimized EOF detection
#                       Optimized Content-length dection (still not good)
# (09/09/06) - v2.0.1 - Added dynamic OS for useragent
#                       Changed default of crlf to 1
#                       Changed default of timeout to 30
#                       Changed from eggdrop sock to TCL sock
# (06/08/06) - v2.0.0 - Changed header/body splitting (supports binary data)
#                       Added content type/charset/length vars
#                       Added content encoding support (automatically convert)
# (03/31/05) - v1.1.2 - Added -host
# (03/04/05) - v1.1.1 - Added -connection, -charset, -language and -accept
#                     - replaced proc isint with string is integer
#
# removed original change log, refer original egghttp
#
##
#
# Usage:
#
# See description before each procedure, and also
# see bottom of script for example usage.
#
# Note: Load this script BEFORE any other script that requires this tcl.
#
##

# Check for this variable to see if this TCL is loaded
set egghttp(version) "2.1.0"

# set OS (for Useragent)
if {[file executable /bin/uname]} {
# sollte unter BSD/Linux/Cygwin etc. funktionieren
	set egghttp(OS) [exec /bin/uname -sr]
} elseif {[file executable uname]} {
	set egghttp(OS) [exec uname -sr]
} elseif {[info exists env(OS)]} {
# sollte unter Windows ohne Cygwin Installation funktionieren
	set egghttp(OS) $env(OS)
} else {
	set egghttp(OS) "Unix compatible"
}

# detect decompression capility
if {[catch {exec unzip -v}]} {
	set egghttp(zip) 0
} else {
	set egghttp(zip) 1
}

if {[catch {exec gunzip --version}]} {
	set egghttp(gzip) 0
} else {
	set egghttp(gzip) 1
}

# regular expression for URL parsing
set re(egghttp) {^(http://)?([^:/]+)(:([0-9]+))?(/.*)?$}

# no stupid try to keep backwards compatible
package require Tcl 8.0

####
#
# Procedure: egghttp:geturl
#
# Description: Used to download the contents of a webpage
#
# Arguments: url        = webpage to download
#            command    = command to execute when transaction is
#                         complete.  This command is called with
#                         one parameter, the sockID
#            options    = -timeout   -> Seconds before conection times out.
#                                       (Default = 60 seconds)
#                         -query     -> Query a webpage script (ie. cgi's)
#                         -headers   -> Send header information to server
#                                       (ie. Cookies)
#                         -protocol  -> Protocol to use (Default = HTTP/1.0)
#                         -useragent -> Useragent to reply with to server
#                                       (Default = Mozilla/5.0)
#                         -crlf      -> 0 or 1, Use CRLF's with query
#                                       (Default = 0, no)
#
# Returns: sockID
#
####

proc egghttp:geturl {url command args} {
	global egghttp
	if {![regexp -nocase $::re(egghttp) $url x protocol server y port path]} {
		return -code error "bogus URL: $url"
	}
	if {[string length $port] == 0} {
		set port 80
	}

	set state(-timeout) 30
	set state(-query) ""
	set state(-headers) ""
	set state(-host) "$server"
	set state(-protocol) "HTTP/1.1"
	set state(-accept) "text/html,text/plain;q=0.9,text/xml;q=0.8,text/*;q=0.7,application/xhtml+xml;q=0.6,application/xml;q=0.5,*;q=0.1"
	set state(-language) "de-de,de;q=0.9,de*;q=0.9,en-us;q=0.7,en;q=0.6,en*;q=0.6,*;q=0.1"
	set state(-charset) "ISO-8859-1,windows-1252;q=0.9,ISO-8859-15;q=0.8,UTF-8;q=0.5,*;q=0.1"
	set state(-useragent) "Tcl http client (TCL [info patchlevel]; egghttp $egghttp(version); de-DE; $egghttp(OS)"
	set state(-crlf) 1
	set state(-encoding) "chunked;q=0.5"
	set state(-myaddr) ${::my-ip}
	
#	if {$egghttp(zip)} {
#		append state(-encoding) ", deflate;q=0.7"
#	}
#	if {$egghttp(gzip)} {
#		append state(-encoding) ", gzip;q=0.9"
#	}
	if {$state(-myaddr) == "" && ${::my-hostname} != ""} {
  		set state(-myaddr) ${::my-hostname}
  	}

	set options {-timeout -query -headers -protocol -accept -language -charset -useragent -crlf -host -encoding -myaddr}
	set usage [join $options ", "]
	foreach {item value} $args {
	  if {[lsearch $options $item] != -1} {
	    if {[info exists state($item)] && [string is integer -strict $state($item)] && ![string is integer -strict $value]} {
	      return -code error "Bad value for $item ($value), must be integer"
	    }
	    set state($item) $value
	  } else {
	    return -code error "Unknown option $item, can be: $usage"
	  }
	}
	set sock [socket -async $server $port]
	fconfigure $sock -blocking 0
	if {$state(-crlf) == 1} {
		fconfigure $sock -translation {auto crlf}
	} elseif {$state(-crlf) == 0} {
		fconfigure $sock -translation {auto lf}
	} else {
		fconfigure $sock -translation binary
	}
	if {$state(-query) == ""} {
		puts $sock "GET $path $state(-protocol)"
		puts $sock "Host: $state(-host):$port"
		puts $sock "Accept: $state(-accept)"
		puts $sock "Accept-Language: $state(-language)"
		puts $sock "Accept-Charset: $state(-charset)"
		puts $sock "Accept-Encoding: $state(-encoding)"
		puts $sock "Connection: close"
		puts $sock "User-Agent: $state(-useragent)"
		if {$state(-headers) != ""} {
			puts $sock "$state(-headers)"
		}
		puts $sock ""
	} else {
		set length [string length $state(-query)]
		puts $sock "POST $path $state(-protocol)"
		puts $sock "Accept: $state(-accept):$port"
		puts $sock "Accept-Language: $state(-language)"
		puts $sock "Accept-Charset: $state(-charset)"
		puts $sock "Accept-Encoding: $state(-encoding)"
		puts $sock "Host: $state(-host)"
		puts $sock "Connection: close"
		puts $sock "User-Agent: $state(-useragent)"
		if {$state(-headers) != ""} {
			puts $sock "$state(-headers)"
		}
		puts $sock "Content-Type: application/x-www-form-urlencoded"
		puts $sock "Content-Length: $length"
		puts $sock ""
		puts $sock "$state(-query)"
	}
	set egghttp($sock,url) "$url"
	set egghttp($sock,error) "Ok"
	set egghttp($sock,command) $command
	set egghttp($sock,code) 0
	set egghttp($sock,length) 0
	set egghttp($sock,wasn) 0
	set egghttp($sock,charset) ""
	set egghttp($sock,input) ""
	set egghttp($sock,crlf) $state(-crlf)
	if {[catch {flush $sock} error]} {
		catch {close $sock}
		set egghttp($sock,error) $error
		catch {eval $command $sock}
		return 1
	}
	set egghttp($sock,timer) [utimer $state(-timeout) "egghttp:timeout $sock"]
	fileevent $sock readable [list egghttp:control $sock]
	return $sock
}

####
#
# Procedure: egghttp:cleanup
#
# Description: Used to clean up variables that are no longer needed
#
# Arguments: sockID     = the sockID of the connection to clean up
#
# Returns: nothing
#
####
proc egghttp:cleanup {sock} {
#  global egghttp
# blah.. would normally just do "array unset egghttp $sock,*"
# but earlier tcl versions don't support it...
#  foreach blah [array names egghttp $sock,*] {
#    catch {unset egghttp($blah)}
#  }
	array unset ::egghttp $sock,*
}

####
#
# Procedure: egghttp:timeout
#
# Description: Used to timeout a connection. Do NOT call this manually
#
# Arguments: sockID     = sockID to timeout
#
# Returns: nothing
#
####
proc egghttp:timeout {sock} {
	global egghttp
	catch {close $sock}
	set egghttp($sock,error) "Timeout or Connection Refused"
	catch {eval $egghttp($sock,command) $sock}
}

####
#
# Procedure: egghttp:data
#
# Description: Used to return the contents of the downloaded page
#
# Arguments: sockID     = sockID of the data to return
#
# Returns: contents of webpage
#
####
proc egghttp:data {sock} {
  global egghttp
  if {[info exists egghttp($sock,body)]} {
    return "$egghttp($sock,body)"
  }
  return ""
}

####
#
# Procedure: egghttp:headers
#
# Description: Used to return the header content of the downloaded page
#
# Arguments: sockID     = sockID of the data to return
#
# Returns: header contents of webpage
#
####
proc egghttp:headers {sock} {
  global egghttp
  if {[info exists egghttp($sock,headers)]} {
    return "$egghttp($sock,headers)"
  }
  return ""
}

####
#
# Procedure: egghttp:errormsg
#
# Description: Used to return any errors while getting page
#
# Arguments: sockID     = sockID of the data to return
#
# Returns: error message, or "Ok" if no error.
#
####
proc egghttp:errormsg {sock} {
  global egghttp
  if {[info exists egghttp($sock,error)]} {
    return "$egghttp($sock,error)"
  }
  return "Ok"
}

####
#
# Procedure: egghttp:code
#
# Description: Used to return the code received from the server while getting page
#
# Arguments: sockID     = sockID of the data to return
#
# Returns: code received by server, or "" if no code was received/found.
#
####
proc egghttp:code {sock} {
  global egghttp
  if {[info exists egghttp($sock,code)]} {
    return "$egghttp($sock,code)"
  }
  return ""
}

####
#
# Procedure: egghttp:control
#
# Description: Used to control incoming traffic from page. Do NOT call
#              this manually.
#
# Arguments: sockID     = sockID of connection
#            input      = incoming data
#
# Returns: 1 to relinquish control, 0 to retain control
#
####
proc egghttp:control {sock} {
	global egghttp
	set endh 0
	set read 0
	set input $egghttp($sock,input)
	set error [fconfigure $sock -error]
	if {![info exists egghttp($sock,body)]} {
		while {[set c [read $sock 1]] != ""} {
			append input $c
			incr read [string bytelength $c]
			if {$c == "\n"} {
				if {$egghttp($sock,wasn)} {
					set endh 1
#					putdcc [hand2idx De_Kus] "End of Header"
					break
				} else {
					set egghttp($sock,wasn) 1
				}
			} elseif {$c != "\r"} {
				set egghttp($sock,wasn) 0
			}
		}
		if {$endh} {
			set header $input
			set egghttp($sock,input) ""
			scan $header "%*s %d" egghttp($sock,code)
			set headertable [lrange [split $header \n] 1 end-2]
			set hlength [llength $headertable]
			for {set i 0} {$i < $hlength} {incr i} {
				set line [split [lindex $headertable $i] :]
				lappend newheader [string tolower [lindex $line 0]] [string trim [join [lrange $line 1 end] :]]
			}
			array set headers [set egghttp($sock,headers) $newheader]
			unset newheader headertable
			if {[info exists headers(content-type)]} {
				set header [split [join [lrange [split $headers(content-type) ;] 1 end] ;] =;]
				foreach {name value} $header {
					if {[string equal -nocase $name charset]} {
						set egghttp($sock,charset) [string map {iso- iso Windows- cp us- {}} [string tolower [string trim $value]]]
						break
					}
				}
			}
			if {[info exists headers(content-length)]} {
				set egghttp($sock,length) $headers(content-length)
			}
			if {[info exists headers(content-encoding)] && ![string equal -nocase $headers(content-encoding) chunked]} {
				set egghttp($sock,crlf) 2
				fconfigure $sock -translation binary
#				putdcc [hand2idx De_Kus] "detected binary data. Content-Encoding: $headers(content-encoding)"
			} elseif {[info exists egghttp($sock,charset)]} {
				if {[lsearch [encoding names] $egghttp($sock,charset)] != -1} {
					fconfigure $sock -encoding $egghttp($sock,charset)
				} else {
					putlog "Egghttp2 warning: Server returned unsupported and unrequested charset '$egghttp($sock,charset)'. Parsing stream as '[encoding system]'"
				}
			}
		} else {
			append egghttp($sock,input) $input
			if {$c == "" && [set error [fconfigure $sock -error]] == ""} {
				return 0
			}
		}
	}
	if {[info exists egghttp($sock,body)]} {
		set bodyl [string bytelength $egghttp($sock,body)]
	} else {
		set bodyl 0
	}
#	putdcc [hand2idx De_Kus] "Read: '$read' EoF: '[eof $sock]'"
	set input [read $sock]
# dont laugh, but this makes trigger eof earlier than it sometimes would
	append input [read $sock]
	set read [string bytelength $input]
	append egghttp($sock,body) $input
	set bodyl [string bytelength $egghttp($sock,body)]
#	putdcc [hand2idx De_Kus] "Sock: '$sock' Input: '[string bytelength $input]' Expected: '$egghttp($sock,length)' EoF: '[eof $sock]' Error: '$error' Blocked: '[fblocked $sock]'"
	if {[eof $sock] || (![fblocked $sock] && $read == 0) || $error != ""} {
		catch {killutimer $egghttp($sock,timer)}
		catch {close $sock}
		if {$egghttp($sock,body) != ""} {
			if {![info exists headers]} {
				array set headers [egghttp:headers $sock]
			}
			if {[info exists headers(content-encoding)] && ![string equal -nocase chunked $headers(content-encoding)]} {
				if {($egghttp(gzip) && [string equal -nocase gzip $headers(content-encoding)]) || ($egghttp(zip) && [string equal -nocase deflate $headers(content-encoding)])} {
# create file in temp directory defined in the eggdrop config
					set file "[string trimright ${::temp-path} /]/egghttp$sock"
					if {[catch {set fd [open $file w]} err]} {
						set egghttp($sock,error) [string trim $err \n]
						if {[catch {close $fd} err]} {
#							putlog "egghttp2: [string trim $err \n]"
							set egghttp($sock,error) [string trim $err \n]
						}
					} else {
# even if the file might be only present temporary, lets ensure privacy!
						file attributes $file -permissions 00600
# lets make sure no encoding screws up the compressed binary data!
						fconfigure $fd -translation binary
						puts -nonewline $fd $egghttp($sock,body)
						catch {close $fd}
						if {[string equal -nocase deflate $headers(content-encoding)]} {
							set fd [open "|unzip -pq <$file" r]
						} else {
							set fd [open "|gunzip -q <$file" r]
						}
#						putdcc [hand2idx De_Kus] unzipping
						fconfigure $fd -blocking 1 -buffering none
						if {[info exists egghttp($sock,charset)] && $egghttp($sock,charset) != "" && ![string equal -nocase [encoding system] $egghttp($sock,charset)]} {
							if {[lsearch [encoding names] $egghttp($sock,charset)] != -1} {
								fconfigure $fd -encoding $egghttp($sock,charset)
							} else {
								putlog "Egghttp2 warning: Server returned unsupported and unrequested charset '$egghttp($sock,charset)'. Threating body as '[encoding system]'"
							}
						}
						set egghttp($sock,body) [read $fd]
						if {[catch {close $fd} err]} {
#						   putlog "egghttp2: [string trim $err \n]"
							set egghttp($sock,error) [string trim $err \n]
						}
					}
#					catch {file delete -force $file}
				} else {
# I know some servers ignore Accept-Encoding HTTP headers, so let's inform the user about it
					set egghttp($sock,error) "Server returned the unsuppprted and/or unrequested Content-Encoding '$headers(content-encoding)'"
				}		
			}
		}
		if {$error != "" && $error != "connection reset by peer"} {
			putlog "Socket '$sock' from egghttp2.tcl closed for '$error'"
			set egghttp($sock,error) $error
		}
		catch {eval $egghttp($sock,command) $sock}
		return 1
	}
	return 0
}

putlog "egghttp2.tcl API v$egghttp(version) by strikelight, modified by De Kus, now loaded."

###
#
# Example 1:
#
# proc connect_callback {sock} {
#   if {[egghttp:errormsg $sock] != "Ok"} {
#      .. whaterever you want to do in case of an error ..
#   }
#   array set headers [egghttp:headers $sock]
#   set buffer [egghttp:data $sock]
#   egghttp:cleanup $sock
#   .. whatever else you want to do with the data ..
# }
#
# set sock [egghttp:geturl www.test.com/ connect_callback]
#
# Example 2: (Query a cgi script)
#
# same proc connect_callback
#
# set sock [egghttp:geturl www.test.com/test.cgi connect_callback -query input=blah]
#
# Example 3: (Send header information, such as cookies)
#
# same proc connect_callback
#
# set sock [egghttp:geturl www.test.com/ connect_callback -headers "Cookie: uNF=unf"]
#
###
